<!DOCTYPE html>
<html{% block iommi_html_tag_attrs %} lang="{{ iommi_language_code }}"{% endblock %}>
<head>
    <title>{% block title %}{{ title }}{% endblock %}</title>
    {% block iommi_head_contents %}{% endblock %}
    {% for asset in assets.values %}
        {% if not asset.in_body %}
            {{ asset }}
        {% endif %}
    {% endfor %}
    {% block iommi_head_contents_last %}{% endblock %}
</head>
<body{% block iommi_body_tag_attrs %}{% if request.iommi_main_menu %} class="sidebar_menu_active"{% endif %}{% endblock %}>
    {% block iommi_main_menu %}
        {% if request.iommi_main_menu %}
            {{ request.iommi_main_menu }}
        {% endif %}
    {% endblock %}

    {% block iommi_top %}{% endblock %}

    {{ container.iommi_open_tag }}
        {% block content %}{% endblock %}
    {{ container.iommi_close_tag }}

    {% block iommi_bottom %}{% endblock %}

    {% for asset in assets.values %}
        {% if asset.in_body %}
            {{ asset }}
        {% endif %}
    {% endfor %}
</body>
</html>
