.iommi_filter .iommi_query_error {
    color: red;
}
.iommi_filter .iommi_query_error.hidden {
    display: none;
}

/* advanced filter */
.iommi_filter .iommi_query_available_fields,
.iommi_filter .iommi_query_available_query_commands {
    display: inline-block;
    vertical-align: top;
    margin-right: 10px;
    max-width: 45%;
}
.iommi_filter .iommi_query_available_fields {
    padding-right: 20px;
}
.iommi_filter .iommi_query_available_fields ul,
.iommi_filter .iommi_query_available_query_commands ul {

}
.iommi_filter .iommi_query_toggle_help,
.iommi_filter .iommi_query_help {
    margin-left: 3px;
}
.iommi_filter .iommi_query_toggle_help {
    cursor: pointer;
    width: 100px;
}
.iommi_filter .iommi_query_toggle_simple_mode {
    float: right;
}
/* /advanced filter */

/* EditTable */
input[data-iommi-edit-table-delete-row-checkbox] {
    display: none;
}
input[data-iommi-edit-table-delete-row-checkbox]+label {
    display: none;
}
[data-iommi-edit-table-delete-with="checkbox"] button[data-iommi-edit-table-delete-row-button] {
    display: none;
}
[data-iommi-edit-table-delete-with="checkbox"] input[data-iommi-edit-table-delete-row-checkbox] {
    display: initial;
}
div[data-iommi-edit-table-delete-with="checkbox"] input[data-iommi-edit-table-delete-row-checkbox]+label {
    display: initial;
}
[data-iommi-type="EditCells"] {
    opacity: 1;
}
[data-iommi-type="EditCells"]:has(input[data-iommi-edit-table-delete-row-checkbox]:checked) {
    opacity: 0.3;
}
.reordering-handle-cell {
	cursor: move;
	cursor: -webkit-grabbing;
    width: 30px;
    text-align: right;
}
.reordering-handle-cell::after {
    content: "||";
}
/* /EditTable */
