<th{{ header.attrs }}>
    <i class="{{ header.column.extra.icon }}"
       onclick="iommi_table_js_select_all(this, {% if header.table.parts.page.is_paginated %}true{% else %}false{% endif %})"></i>
</th>
